#ifndef UNARYFUNCTION_H
#define UNARYFUNCTION_H

#include <list>
#include <algorithm>

template <typename T>
class UnaryFunction
{
public:
    UnaryFunction(T amount): m_amount(amount) {}
    virtual T operator() (T argument) const = 0;
    // Musi mit virtualni destruktor, jinak se spravne neuvolni pamet
    virtual ~UnaryFunction() {};
protected:
    T m_amount;
};

template <typename T>
void delete_ptr(UnaryFunction<T>* ptr)
{
    delete ptr;
}

template <typename T, size_t N=0>
class FunctionCompositor {
public:
    FunctionCompositor() {}
    ~FunctionCompositor() { std::for_each(m_functions.begin(), m_functions.end(), delete_ptr<T>); }

    void appendFunction(UnaryFunction<T>* fn) { m_functions.push_back(fn); }

    T operator() (T arg)
    {
        typedef typename std::list<UnaryFunction<T>*>::iterator iterator;
        for (iterator it = m_functions.begin(); it != m_functions.end(); ++it)
        {
            UnaryFunction<T>* fnc = *it; // Get particular function
            arg = (*fnc)(arg); // Apply it on argument
        }
        return arg;
    }
private:
    std::list<UnaryFunction<T>*> m_functions;
};


// Jednodussi varinta

template <typename T>
class FunctionCompositor<T*,0> {
public:
    FunctionCompositor() {}
    ~FunctionCompositor() { std::for_each(m_functions.begin(), m_functions.end(), delete_ptr<T>); }

    void appendFunction(UnaryFunction<T>* fn) { m_functions.push_back(fn); }

    T* operator() (T* arg)
    {
        typedef typename std::list<UnaryFunction<T>*>::iterator iterator;
        for (iterator it = m_functions.begin(); it != m_functions.end(); ++it)
        {
            UnaryFunction<T>* fnc = *it; // Get particular function
            *arg = (*fnc)(*arg); // Apply it on argument
        }
        return arg;
    }
private:
    std::list<UnaryFunction<T>*> m_functions;
};

template <>
class FunctionCompositor<unsigned int*,0> {
public:
    FunctionCompositor() {}
    ~FunctionCompositor() { std::for_each(m_functions.begin(), m_functions.end(), delete_ptr<unsigned int>); }

    void appendFunction(UnaryFunction<unsigned int>* fn) { m_functions.push_back(fn); }

    unsigned int* operator() (unsigned int* arg)
    {
        typedef std::list<UnaryFunction<unsigned int>*>::iterator iterator;
        for (iterator it = m_functions.begin(); it != m_functions.end(); ++it)
        {
            UnaryFunction<unsigned int>* fnc = *it; // Get particular function
            *arg = (*fnc)(*arg); // Apply it on argument
        }
        return arg;
    }
private:
    std::list<UnaryFunction<unsigned int>*> m_functions;
};

// Slozitejsi varianta

template <typename T, size_t N>
class FunctionCompositor<T*,N> {
public:
    FunctionCompositor() {}
    ~FunctionCompositor() { std::for_each(m_functions.begin(), m_functions.end(), delete_ptr<T>); }

    void appendFunction(UnaryFunction<T>* fn) { m_functions.push_back(fn); }

    T* operator() (T* arg)
    {
        typedef typename std::list<UnaryFunction<T>*>::iterator iterator;
        for (iterator it = m_functions.begin(); it != m_functions.end(); ++it)
        {
            UnaryFunction<T>* fnc = *it; // Get particular function
            for(size_t i = 0; i < N; i++)
            {
                arg[i] = (*fnc)(arg[i]); // Apply it on argument
            }
        }
        return arg;
    }
private:
    std::list<UnaryFunction<T>*> m_functions;
};

template <size_t N>
class FunctionCompositor<unsigned int*,N> {
public:
    FunctionCompositor() {}
    ~FunctionCompositor() { std::for_each(m_functions.begin(), m_functions.end(), delete_ptr<unsigned int>); }

    void appendFunction(UnaryFunction<unsigned int>* fn) { m_functions.push_back(fn); }

    unsigned int* operator() (unsigned int* arg)
    {
        typedef std::list<UnaryFunction<unsigned int>*>::iterator iterator;
        for (iterator it = m_functions.begin(); it != m_functions.end(); ++it)
        {
            UnaryFunction<unsigned int>* fnc = *it; // Get particular function
            for(size_t i = 0; i < N; i++)
            {
                arg[i] = (*fnc)(arg[i]); // Apply it on argument
            }
        }
        return arg;
    }
private:
    std::list<UnaryFunction<unsigned int>*> m_functions;
};

#endif
